//#include "stdint.h"
#include "DDModbusMaster.h"
#include "HardwareSerial.h"


// DD - 2015-05-06
// Optionally debugout an exception value -> processReply()
// "Serial" as console is hardcoded here.

// DD - 2015-06-04
// - enhanced timing in waiting_for_reply.  
// - methods added for activation and deactivation of calls (packets) -> activateCall(), deactivateCall()
// - makeCall() included; 
// - version properties added
// - exception handling added to processError()

// DD - 2015-06-30
// Optionally omit the TxEnablePin, when communicating via RS232 rather than EIA-485.
//
// In order to do so, send 0 for the _TxEnablePin in modbus_configure().
// Paul Alting, modified so that if set to 0xff, then no TxEnablePin is set or used.
// Pin 0 is a valid pin on the Arduino, so best not to block the usage of that.

/*
	Copyright (c) Frank Dietrich Datentechnik 2015
    frank.dietrich@dd-tech.de
	
    This is a partial rewrite and enhanced Version of 
    SimpleModbusMasterV2rev2 by Juan Bester
    -> https://code.google.com/p/simple-modbus/
    
    As the handling now differs from Juan's version I
    renamed it to DDModbusMaster so that both versions
    can coexist. 
        
    
    Todo: 
    
    - check calculations of frame and inter frame char delays?
	- eventually add current callState to connecton property. 
	  callStates could be: pending (wait4pickup), wait4response, success, fail... 
	  
    - nach einem eventuellen Timeout aller Befehle ggf. dennoch alle x Sekunden einen 
       Reaktivierungsversuch starten. Andernfalls sind z.B. alle reinen Statusabfragen dauerhaft
       tot, auch, wenn ggf. später wieder angeschaltet wird. 
       oder einen keep alive call der kein retry limit hat
       Eventuell eine Reset-Funktion einbauen.   
    
	

    Done:

	- Laut Spezifikation ist Turnaround-Delay nur bei Broadcast notwendig - kontrollieren
    - Callback-Method onCallTerminated( Packet, Success, Exception code )  
    - Extended functionality of the .connection - property ausweiten. 
	  Its bits now have different tasks: 
        - active / inactive  (bit 0). If set, the packet is currently active.
        - one time / continuous activation  (bit 2). 
		  If set, the packet will be called only once when the active bit (bit 0) is set high.
        - deferred until next round vs. run as soon as possible. (bit 1) .
		  If set, the packet will be set active when the packet loop counter restarts at 0.
		  Important if commands depend on other environmental settings.
		  Example: Setting a new frequency on a VFD causes an exception if it not in NET mode.
		           Thus prior to setting a new frequency, NET mode has to be checked or set 
				   in case the user deactivated it via the input panel.
    - Currently waiting_for_reply() forces a delay of the maximum inter char wait time 
	  after each character it receives from the buffer. That surely could be optimized.
	- processError() could also report crc or invalid function errors.  
    
*/



/*

unsigned char state;
unsigned char retry_count;
unsigned char TxEnablePin;

// frame[] is used to receive and transmit packages. 
// The maximum number of bytes in a modbus packet is 256 bytes
// This is limited to the serial buffer of 64 bytes
unsigned char frame[BUFFER_SIZE]; 
unsigned char buffer;
long timeout; // timeout interval
long polling; // turnaround delay interval
unsigned int T1_5; // inter character time out in microseconds
unsigned int frameDelay; // frame time out in microseconds
long delayStart; // init variable for turnaround and timeout delay
unsigned int total_no_of_packets; 
Packet* packetArray; // packet starting address
Packet* packet; // current packet
unsigned int* register_array; // pointer to masters register array
HardwareSerial* ModbusPort;

Callback CallbackFunction ;     

// function definitions
void idle();
void constructPacket();
unsigned char construct_F15();
unsigned char construct_F16();
void waiting_for_reply();
void processReply();
void initTurnaround();
void waiting_for_turnaround();
void process_F1_F2();
void process_F3_F4();
void process_F5_F6_F15_F16();
void processError();
void processSuccess();
unsigned int calculateCRC(unsigned char bufferSize);
void sendPacket(unsigned char bufferSize);

*/


// Modbus Master State Machine
DDModbusMaster :: DDModbusMaster(){
}

DDModbusMaster :: ~DDModbusMaster(){
}


///////////////////////////////////
// Public methods / functions ////
///////////////////////////////////

void DDModbusMaster :: modbus_update() {
    switch (state) {
        case IDLE:
            idle();
            break;
        
        case WAITING_FOR_REPLY:
            waiting_for_reply();
            break;
        
        case WAITING_FOR_TURNAROUND:
            waiting_for_turnaround();
            break;
    }

}


void DDModbusMaster :: modbus_configure( HardwareSerial* SerialPort,
                        long baud,
                        unsigned char byteFormat,
                        long _timeout, 
                        long _polling, 
                        unsigned char _retry_count, 
                        unsigned char _TxEnablePin, 
                        Packet* _packets, 
                        unsigned int _total_no_of_packets,
                        unsigned int* _register_array, 
                        Callback _callbackFn ) {
                         
    // Modbus states that a baud rate higher than 19200 must use a fixed 750 us 
    // for inter character time out and 1.75 ms for a frame delay for baud rates
    // below 19200 the timing is more critical and has to be calculated.
    // E.g. 9600 baud in a 11 bit packet is 9600/11 = 872 characters per second
    // In milliseconds this will be 872 characters per 1000ms. So for 1 character
    // 1000ms/872 characters is 1.14583ms per character and finally modbus states
    // an inter-character must be 1.5T or 1.5 times longer than a character. Thus
    // 1.5T = 1.14583ms * 1.5 = 1.71875ms. A frame delay is 3.5T.
    // Thus the formula is T1.5(us) = (1000ms * 1000(us) * 1.5 * 11bits)/baud
    // 1000ms * 1000(us) * 1.5 * 11bits = 16500000 can be calculated as a constant
    
    if (baud > 19200) {
        T1_5 = 750; 
    } else {
        T1_5 = 16500000/baud; // 1T * 1.5 = T1.5
    }    
        
    /* The modbus definition of a frame delay is a waiting period of 3.5 character times
         between packets. This is not quite the same as the frameDelay implemented in
         this library but does benifit from it.
         The frameDelay variable is mainly used to ensure that the last character is 
         transmitted without truncation. A value of 2 character times is chosen which
         should suffice without holding the bus line high for too long.*/
         
    frameDelay = T1_5 * 2; 
    
    // initialize
    state = IDLE;
    timeout = _timeout;
    polling = _polling;
    retry_count = _retry_count;
    TxEnablePin = _TxEnablePin;
    total_no_of_packets = _total_no_of_packets;
    packetArray = _packets;
    register_array = _register_array;
    
    ModbusPort = SerialPort;
    (*ModbusPort).begin(baud, byteFormat);
    

	// 2015-06-30
	// skip when not needed
//	if (TxEnablePin > 0)
    if (TxEnablePin != 0xff)
	{
		pinMode(TxEnablePin, OUTPUT);
		digitalWrite(TxEnablePin, LOW);
	}
	
    
    // DD 2015-05-09
    // Callbackfunction that will report back to main loop
    // whenever a call is terminated either successfully or
    // with failure.
    CallbackFunction = _callbackFn;
    
} 


void DDModbusMaster :: modbus_construct( Packet *_packet, 
                        unsigned char id, 
                        unsigned char function, 
                        unsigned int address,     
                        unsigned int data, 
                        unsigned int local_start_address,
						uint8_t connection ){
                        
    _packet->id = id;
    _packet->function = function;
    _packet->address = address;
    _packet->data = data;
    _packet->local_start_address = local_start_address;
    
    // DD- 2015-05-14
    // same value, but make clear, we're only using the 
    // the last bit as an active/inactive flag
    //_packet->connection = 1;
    _packet->connection = connection; // bit(0);
    
}


unsigned int DDModbusMaster :: calculateCRC(unsigned char bufferSize) {
    unsigned int temp, temp2, flag;
    temp = 0xFFFF;

    for (unsigned char i = 0; i < bufferSize; i++) {
        temp = temp ^ frame[i];
        
        for (unsigned char j = 1; j <= 8; j++) {
            flag = temp & 0x0001;
            temp >>= 1;
            if (flag) {
                temp ^= 0xA001;
            }    
        }
    }
    
    // Reverse byte order. 
    temp2 = temp >> 8;
    temp = (temp << 8) | temp2;
    temp &= 0xFFFF;
    // the returned value is already swapped
    // crcLo byte is first & crcHi byte is last
    
    return temp; 
}



// initiate a one time deferred call 
void DDModbusMaster :: makeCall( Packet* _packet, unsigned int _packetValue ){
	register_array[_packet->local_start_address] = _packetValue;
    _packet->connection  |= (bit(2) | bit(1));  
}

void DDModbusMaster :: makeCall( uint8_t _packetIndex, unsigned int _packetValue ){
		makeCall( &packetArray[_packetIndex], _packetValue );
}

// activate or deactivate calls
void DDModbusMaster :: activateCall( Packet* _packet ){
	_packet->connection |= B00000001; 
}
void DDModbusMaster :: activateCall( uint8_t _packetIndex ){
	activateCall( &packetArray[_packetIndex] );
}
void DDModbusMaster :: deactivateCall( Packet* _packet ){
	_packet->connection &= ~B00000001; 
}	
void DDModbusMaster :: deactivateCall( uint8_t _packetIndex ){
	deactivateCall( &packetArray[_packetIndex] );
}


// Return the time taken between two microsecond timestamps. 
// respects timer overflow every ~ 70min
// t2 defaults to current micros() setting 
unsigned long DDModbusMaster :: usTimeSpan( unsigned long t1, unsigned long t2){

    if (t2 < t1){
        // Timer-overflow
        return  US_TMRMAX - t1 + t2;      
    } else {
        return t2 - t1;
    }
}



///////////////////////////////////
// Private methods / functions ////
///////////////////////////////////


void DDModbusMaster :: idle() {
	static unsigned int packet_index;
	unsigned int failed_connections = 0;
	unsigned char current_connection;
	do {
		if (packet_index == total_no_of_packets)            // wrap around to the beginning
		{
			packet_index = 0;
		}

		// activate deferred commands
		if (packet_index == 0)
		{
			for (uint8_t i = 0; i < total_no_of_packets; i++)
			{
				if ( packetArray[i].connection & (1 << 1) )
				{
					packetArray[i].connection &= ~(1<<1);
					packetArray[i].connection |= (1<<0);
				}
			}
		}

		packet = &packetArray[packet_index];		// proceed to the next packet
// get the current connection status
// DD - 2015-05-14
// Only the last bit used for active/inactive detection
// current_connection = packet->connection;
		current_connection = (packet->connection & bit(0));

		if (!current_connection)	// If all the connection attributes are false return immediately to the main sketch
		{
			if (++failed_connections == total_no_of_packets)
			{
				return;
			}
		}
	packet_index++;					// if a packet has no connection get the next one
	} while (!current_connection);
	constructPacket();
}

void DDModbusMaster :: constructPacket() {     
    packet->requests++;
    
    frame[0] = packet->id;
    frame[1] = packet->function;
    frame[2] = packet->address >> 8; // address Hi
    frame[3] = packet->address & 0xFF; // address Lo
    
    // For functions 1 & 2 data is the number of points
    // For function 5 data is either ON (0xFF00) or OFF (0x0000)
    // For function 6 data is exactly that, one register's data
    // For functions 3, 4 & 16 data is the number of registers
    // For function 15 data is the number of coils
    
    // The data attribute needs to be intercepted by F5 & F6 because these requests
    // include their data in the data register and not in the masters array
    if (packet->function == FORCE_SINGLE_COIL || packet->function == PRESET_SINGLE_REGISTER) {
        packet->data = register_array[packet->local_start_address]; // get the data
    }
    
    frame[4] = packet->data >> 8; // MSB
    frame[5] = packet->data & 0xFF; // LSB    eigentlich 0x00FF
    
    unsigned char frameSize;
    
    // construct the frame according to the modbus function  
    if (packet->function == PRESET_MULTIPLE_REGISTERS) {
        frameSize = construct_F16();
        
    } else if (packet->function == FORCE_MULTIPLE_COILS) {
        frameSize = construct_F15();
        
    } else {         // else functions 1,2,3,4,5 & 6 is assumed. They all share the exact same request format.
        frameSize = 8; // the request is always 8 bytes in size for the above mentioned functions.
    }
        
    unsigned int crc16 = calculateCRC(frameSize - 2);    
      frame[frameSize - 2] = crc16 >> 8; // split crc into 2 bytes
    frame[frameSize - 1] = crc16 & 0xFF;
    
    sendPacket(frameSize);

    state = WAITING_FOR_REPLY; // state change
    
    // if broadcast is requested (id == 0) for function 5,6,15 and 16 then override 
    // the previous state and force a success since the slave wont respond
    if (packet->id == BROADCAST_ID) {
        processSuccess();
    }        
}


unsigned char DDModbusMaster :: construct_F15() {
    // function 15 coil information is packed LSB first until the first 16 bits are completed
    // It is received the same way..
    unsigned char no_of_registers = packet->data / 16;
    unsigned char no_of_bytes = no_of_registers * 2; 
    
    // if the number of points dont fit in even 2byte amounts (one register) then use another register and pad 
    if (packet->data % 16 > 0) {
        no_of_registers++;
        no_of_bytes++;
    }
    
    frame[6] = no_of_bytes;
    unsigned char bytes_processed = 0;
    unsigned char index = 7; // user data starts at index 7
    unsigned int temp;
    
    for (unsigned char i = 0; i < no_of_registers; i++)  {
        temp = register_array[packet->local_start_address + i]; // get the data
        frame[index] = temp & 0xFF; 
        bytes_processed++;
     
        if (bytes_processed < no_of_bytes) {
            frame[index + 1] = temp >> 8;
            bytes_processed++;
            index += 2;
        }
    }
    
    unsigned char frameSize = (9 + no_of_bytes); // first 7 bytes of the array + 2 bytes CRC + noOfBytes 
    return frameSize;
}

unsigned char DDModbusMaster :: construct_F16() {
    unsigned char no_of_bytes = packet->data * 2; 
    
    // first 6 bytes of the array + no_of_bytes + 2 bytes CRC 
    frame[6] = no_of_bytes; // number of bytes
    unsigned char index = 7; // user data starts at index 7
    unsigned char no_of_registers = packet->data;
    unsigned int temp;
        
    for (unsigned char i = 0; i < no_of_registers; i++) {
        temp = register_array[packet->local_start_address + i]; // get the data
        frame[index] = temp >> 8;
        index++;
        frame[index] = temp & 0xFF;
        index++;
    }
    
    unsigned char frameSize = (9 + no_of_bytes); // first 7 bytes of the array + 2 bytes CRC + noOfBytes 
    return frameSize;
}


void DDModbusMaster :: initTurnaround(){
  // DD 2015-05-08
  // init TURNAROUND delay, but only
  // for broadcast-messages 
  // -> processError() / processSuccess()
   
  if (packet->id != BROADCAST_ID){
       // has this been a broadcast message,
       // then give the clients some time to
       // process 
       state = WAITING_FOR_TURNAROUND;
       delayStart = millis(); // start the turnaround delay
       
   }  else {
       // no broadcast message, then return to IDLE mode
       state = IDLE;
   }  

}

void DDModbusMaster :: waiting_for_turnaround() {
    if ((millis() - delayStart) > polling) {
        state = IDLE;
    }    
}

// get the serial data from the buffer
void DDModbusMaster :: waiting_for_reply() {

    if ( (*ModbusPort).available() ) {        
        // is there something to check?    
        unsigned char overflowFlag = 0;
        buffer = 0;
        
		// DD 2015-06-04
		// try not to force a delay after each
		// character. 
		unsigned long tLastChar = micros();
				
        // while ((*ModbusPort).available()) {
		while ( (*ModbusPort).available() || usTimeSpan( tLastChar ) <= T1_5 ){
            // The maximum number of bytes is limited to the serial buffer size 
            // of BUFFER_SIZE. If more bytes is received than the BUFFER_SIZE the 
            // overflow flag will be set and the serial buffer will be read until
            // all the data is cleared from the receive buffer, while the slave is 
            // still responding.
            
			if ( (*ModbusPort).available() ) { 
			
				if (overflowFlag) {
					(*ModbusPort).read();
				} else {
					if (buffer == BUFFER_SIZE) {
						overflowFlag = 1;
					}
					
					frame[buffer] = (*ModbusPort).read();
					buffer++;
				}
				
				tLastChar = micros();				
			}	
				
            // This is not 100% correct but it will suffice.
            // worst case scenario is if more than one character time expires
            // while reading from the buffer then the buffer is most likely empty
            // If there are more bytes after such a delay it is not supposed to
            // be received and thus will force a frame_error.
            // delayMicroseconds(T1_5); // inter character time out
        }
            
        // The minimum buffer size from a slave can be an exception response of
        // 5 bytes. If the buffer was partially filled set a frame_error.
        // The maximum number of bytes in a modbus packet is 256 bytes.
        // The serial buffer limits this to 64 bytes.
    
        if ((buffer < 5) || overflowFlag) {
            processError(); 
        
            // Modbus over serial line datasheet states that if an unexpected slave 
            // responded the master must do nothing and continue with the time out.
            // This seems silly cause if an incorrect slave responded you would want to
            // have a quick turnaround and poll the right one again. If an unexpected 
            // slave responded it will most likely be a frame error in any event

        } else if (frame[0] != packet->id) { // check id returned
            processError();
        } else {
            processReply();
        }
            
    } else if ((millis() - delayStart) > timeout) {    // check timeout
        processError();
        state = IDLE; //state change, override processError() state
    }
}



void DDModbusMaster :: processReply() {
    // combine the crc Low & High bytes
    unsigned int received_crc = ((frame[buffer - 2] << 8) | frame[buffer - 1]); 
    unsigned int calculated_crc = calculateCRC(buffer - 2);
    
    // verify checksum
    if (calculated_crc == received_crc) { 
        // To indicate an exception response a slave will 'OR' 
        // the requested function with 0x80 
        
        // extract 0x80
        if ((frame[1] & 0x80) == 0x80) {    
            packet->exception_errors++;
            
            // 2015-05-06
            // Print out the exception response?
            // Third byte will be an errorcode of 8bit length
            /*
            if (DEBUG_EXCEPTION && buffer >= 3 && Serial){
                Serial.println("\n==============================");
                Serial.println("AN EXCEPTION OCCURRED: ");
                Serial.print("Message: ");
                Serial.println( frame[2] );
                Serial.println("==============================\n");
            }
            */
			
            processError( errException, frame[2] );
            
        } else {
            // No exception; process normal
            switch (frame[1]) { // check function returned
            
                case READ_COIL_STATUS:
        
                case READ_INPUT_STATUS:
                    process_F1_F2();
                    break;
                    
                case READ_INPUT_REGISTERS:
                
                case READ_HOLDING_REGISTERS:
                    process_F3_F4();
                    break;
                
                case FORCE_SINGLE_COIL:
                
                case PRESET_SINGLE_REGISTER:
                
                case FORCE_MULTIPLE_COILS:
                
                case PRESET_MULTIPLE_REGISTERS:
                    process_F5_F6_F15_F16();
                    break;
                
                default: // illegal function returned
                    processError( errIllegalFunction, frame[1] );
                    break;
            }
        }
        
    } else {     
        // checksum failed
        processError(errCRCMismatch, 0);
    }
}



void DDModbusMaster :: process_F1_F2() {
    // packet->data for function 1 & 2 is actually the number of boolean points
    unsigned char no_of_registers = packet->data / 16;
    unsigned char number_of_bytes = no_of_registers * 2; 
    
    // if the number of points dont fit in even 2byte amounts (one register) then use another register and pad 
    if (packet->data % 16 > 0) {
        no_of_registers++;
        number_of_bytes++;
    }
             
    // check number of bytes returned         
    if (frame[2] == number_of_bytes) {      
  
        unsigned char bytes_processed = 0;
        unsigned char index = 3; // start at the 4th element in the frame and combine the Lo byte  
        unsigned int temp;
        
        for (unsigned char i = 0; i < no_of_registers; i++) {
            temp = frame[index]; 
            bytes_processed++;
            
            if (bytes_processed < number_of_bytes) {
                temp = (frame[index + 1] << 8) | temp;
                bytes_processed++;
                index += 2;
            }
            
            register_array[packet->local_start_address + i] = temp;
        }
    
        processSuccess(); 
        
    } else {
        // incorrect number of bytes returned 
        processError(errByteCountMismatch);
    }
}



void DDModbusMaster :: process_F3_F4() {
    // check number of bytes returned - unsigned int == 2 bytes
    // data for function 3 & 4 is the number of registers
    if (frame[2] == (packet->data * 2)) {
        unsigned char index = 3;
        
        for (unsigned char i = 0; i < packet->data; i++) {
            // start at the 4th element in the frame and combine the Lo byte 
            register_array[packet->local_start_address + i] = (frame[index] << 8) | frame[index + 1]; 
            index += 2;
        }
        
        processSuccess(); 
        
    } else {
        // incorrect number of bytes returned  
        processError(errByteCountMismatch);  
    }    
}


void DDModbusMaster :: process_F5_F6_F15_F16() {
    // The repsonse of functions 5,6,15 & 16 are just an echo of the query
    unsigned int recieved_address = ((frame[2] << 8) | frame[3]);
    unsigned int recieved_data = ((frame[4] << 8) | frame[5]);
        
    if ((recieved_address == packet->address) && (recieved_data == packet->data)) {
        processSuccess();
    } else {
        processError(errWrongEcho);
    }
}

void DDModbusMaster :: processError(errorTypes tErrorType, uint8_t tErrorCode) {
    packet->retries++;
    packet->failed_requests++;
        
    // if the number of retries have reached the max number of retries 
    // allowable, stop requesting the specific packet
    if (packet->retries == retry_count) {
        // only lay last bit
        // packet->connection = 0;
        packet->connection &= ~(bit(0));
        packet->retries = 0;
    }


    // DD 2015-05-08
    // If I got the specs right, TURNAROUND applies 
    // to broadcasting messages only. 
    // Created initTurnaround() for that.
    //
    // state = WAITING_FOR_TURNAROUND;
    // delayStart = millis(); // start the turnaround delay
  
    initTurnaround();       

    // DD 2015-05-09
    // report back to main loop
    if (CallbackFunction){  
      CallbackFunction( packet, tErrorType, tErrorCode ); 
    }  

}


void DDModbusMaster :: processSuccess() {
    packet->successful_requests++; // transaction sent successfully
    packet->retries = 0; // if a request was successful reset the retry counter
    
    // DD - 2015-05-14
    // one time call only?
    // check second bit. if set, lay last bit
    if (packet->connection & bit(2)){
        packet->connection &= ~(bit(0));
    }

    // DD 2015-05-08
    // same as in processError(); 
    // turnaround only on broadcast messages
    // state = WAITING_FOR_TURNAROUND;
    // delayStart = millis(); // start the turnaround delay
    
    initTurnaround();

    // DD 2015-05-09
    // report back to main loop
    if (CallbackFunction){  
      CallbackFunction( packet, errSuccess, 0 ); 
    }  
}
  

void DDModbusMaster :: sendPacket(unsigned char bufferSize) {
	// DD 2015-06-30 
	// optionally omit TxEnablePin 
	//
    if (TxEnablePin != 0xff)
	    digitalWrite(TxEnablePin, HIGH);
//	txEnable(HIGH);
        
    for (unsigned char i = 0; i < bufferSize; i++) {
        (*ModbusPort).write(frame[i]);
    }
        
    (*ModbusPort).flush();
    
    delayMicroseconds(frameDelay);
    
    // digitalWrite(TxEnablePin, LOW);
    if (TxEnablePin != 0xff)
	    digitalWrite(TxEnablePin, LOW);
//	txEnable(LOW);
        
    delayStart = millis(); // start the timeout delay    
}


void DDModbusMaster :: txEnable( uint8_t state ){
	// set TxEnablePin HIGH or LOW if being used.
	if (TxEnablePin > 0){
		digitalWrite( TxEnablePin, state );
	}	
}
